/*
 * input.cpp
 *
 *  Created on: 2021年12月4日
 *      Author: pengzc
 */

#include "input.h"
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <math.h>
#include <sys/errno.h>
#include <string.h>
#include <string>
#include "utils/Log.h"
#include <assert.h>
#include <poll.h>

Point from_point_string(const std::string& point) {
  int x = 0;
  int y = 0;
  sscanf(point.c_str(), "%d,%d", &x, &y);
  return Point{ x, y};
}

std::string to_event_string(const struct input_event& e) {
  char buf[64];
  snprintf(buf, int(sizeof(buf)), "%ld %ld %u %u %d",
       e.time.tv_sec, e.time.tv_usec, e.type, e.code, e.value);
  return buf;
}

struct input_event from_event_string(const std::string& str) {
  struct input_event e;
  int ret = sscanf(str.c_str(), "%ld %ld %u %u %d", &e.time.tv_sec,
      &e.time.tv_usec, &e.type, &e.code, &e.value);
  if (ret != 5) {
    LOGE("invalid event string: '%s'", str.c_str());
  }
  return e;
}

static int inject_event(int fd, uint16_t type, uint16_t code, int32_t value) {
  if (fd < 0) {
    LOGE("invalid file descriptor");
    return -1;
  }
  struct input_event event;
  event.type = type;
  event.code = code;
  event.value = value;

  gettimeofday(&event.time, 0);

  if (write(fd, &event, int(sizeof(struct input_event))) < 0) {
      LOGE("fd %d report key error %s!\n", fd, strerror(errno));
      return -1;
  }
  return 0;
}

/**
 * 按下
 */
static int inject_event_down(int fd, int x, int y) {
  //down
  CHECK_FAILURE(inject_event(fd, EV_ABS, ABS_X, x));
  CHECK_FAILURE(inject_event(fd, EV_ABS, ABS_Y, y));
  CHECK_FAILURE(inject_event(fd, EV_ABS, ABS_PRESSURE, Input::kPressureTime));
  CHECK_FAILURE(inject_event(fd, EV_KEY, BTN_TOUCH, 1));
  CHECK_FAILURE(inject_event(fd, EV_SYN, EV_SYN, EV_SYN));
  return 0;
}

//移动
static int inject_event_move(int fd, int x, int y) {
  CHECK_FAILURE(inject_event(fd, EV_ABS, ABS_X, x));
  CHECK_FAILURE(inject_event(fd, EV_ABS, ABS_Y, y));
  CHECK_FAILURE(inject_event(fd, EV_SYN, EV_SYN, EV_SYN));
  return 0;
}

inline int distance(int x1, int y1, int x2, int y2) {
  return sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));
}

static int do_inject_event_move(int fd, int x1, int y1, int x2, int y2) {
  if (distance(x1, y1, x2, y2) < 5) {
    return 0;
  }
  int cx = (x2 + x1) / 2;
  int cy = (y2 + y1) / 2;
  CHECK_FAILURE(do_inject_event_move(fd, x1, y1, cx, cy));
  CHECK_FAILURE(inject_event_move(fd, cx, cy));
  usleep(1000 * Input::kMoveDelayTime);
  CHECK_FAILURE(do_inject_event_move(fd, cx, cy, x2, y2));
  return 0;
}

//抬起
static int inject_event_up(int fd, int x, int y) {
  CHECK_FAILURE(inject_event(fd, EV_ABS, ABS_PRESSURE, 0));
  CHECK_FAILURE(inject_event(fd, EV_KEY, BTN_TOUCH, 0));
  CHECK_FAILURE(inject_event(fd, EV_SYN, EV_SYN, EV_SYN));
  return 0;
}

Input::Input(const std::string& dev) {
  // TODO 自动生成的构造函数存根
  dev_ = dev;
  fd_ = open(dev.c_str(), O_RDWR);
  if (fd_ < 0) {
    LOGE("failed to open %s, %s", dev.c_str(), strerror(errno));
  }
}

Input::~Input() {
  if (fd_ >= 0) {
    close(fd_);
  }
}



int Input::Tap(const std::list<Point>& points, int interval) {
  for (auto it = points.begin(); it != points.end(); ++it) {
    if (interval > 0 && it != points.begin()) {
      usleep(1000 * interval);
    }
    CHECK_FAILURE(Tap(it->x, it->y));
  }
  return 0;
}

int Input::Tap(int x, int y) {
  LOGD("\t tap   (%d,%d)", x, y);
  CHECK_FAILURE(inject_event_down(fd_, x, y));
  CHECK_FAILURE(inject_event_up(fd_, x, y));
  return 0;
}

int Input::Swipe(int x1, int y1, int x2, int y2) {
  LOGD("\t swipe (%d,%d) => (%d,%d)", x1, y1, x2, y2);
  CHECK_FAILURE(inject_event_down(fd_, x1, y1));
  CHECK_FAILURE(do_inject_event_move(fd_, x1, y1, x2, y2));
  CHECK_FAILURE(inject_event_up(fd_, x2, y2));
  return 0;
}

/**
 * -1 error
 * 0 timeout
 * >0 get event
 */
inline int readable(int fd, int timeout_ms) {
  int ret = -1;
  struct pollfd pfd;
  memset(&pfd, 0, sizeof(pfd));
  pfd.fd = fd;
  pfd.events = POLLIN;

  // retry again for EINTR
  for (int i = 0; i < 2; i++) {
    ret = ::poll(&pfd, 1, timeout_ms);
    if (-1 == ret && EINTR == errno)
      continue;
    break;
  }
  return ret;
}

int Input::Wait(struct input_event* buf, int timeout_ms) {
  assert(buf != NULL && "buf must not be null");
  if (fd_ < 0) {
    LOGE("invalid file descriptor");
    return -1;
  }
  int ret = readable(fd_, timeout_ms);
  if (ret <= 0) {
    return ret;
  }
  if (read(fd_, buf, int(sizeof(struct input_event)))
      != int(sizeof(struct input_event))) {
    LOGE("read event failed, %s(%d)\n", strerror(errno), errno);
    return -2;
  }
  return 1;
}

int Input::Inject(const struct input_event& e) {
  if (fd_ < 0) {
    LOGE("invalid file descriptor");
    return -1;
  }
  if (write(fd_, &e, int(sizeof(struct input_event))) < 0) {
      LOGE("fd %d report key error %s!\n", fd_, strerror(errno));
      return -1;
  }
  return 0;
}
