#include <unistd.h>
#include <signal.h>
#include <iostream>
#include <list>
#include "clipp.h"
#include "input.h"
#include "utils/Log.h"
#include <assert.h>

static bool interrupt = false;

static void interrupt_handler(int sig) {
  interrupt = true;
}
enum class Command {
  Tap,
  Swipe,
  Record,
  Playback,
  Monkey,
  Help,
};
struct Args {
  std::string device;
  Command selected_cmd = Command::Help;
  std::string record_file;
  int number_of_times = 1;
  Point p1 = {0, 0};
  Point p2 = {0, 0};
  bool no_swipe;
  int speed = 1;

  int screen_width;
  int screen_height;
};
static Args args;

static int DoMonkey(Input& input, const Args& args) {
  int ret = 0;
  switch (random() % 3) {
    case 0:
    case 1:
    {
      int x = random() % args.screen_width;
      int y = random() % args.screen_height;
      ret = input.Tap(x, y);
    }
      break;
    case 2: {
      if (args.no_swipe) {
        break;
      }
      int x1 = random() % args.screen_width;
      int y1 = random() % args.screen_height;
      int x2 = random() % args.screen_width;
      int y2 = random() % args.screen_height;
      ret = input.Swipe(x1, y1, x2, y2);
    }
      break;
  }
  return ret;
}

static int Monkey(const Args& args) {
  Input input(args.device);
  srand(time(NULL));
  while (!interrupt) {
    CHECK_FAILURE(DoMonkey(input, args));
  }
  return 0;
}

static int AppendFile(const std::list<struct input_event>& list,
    const std::string& file) {
  if (list.empty()) {
    return 0;
  }
  FILE* fp = fopen(file.c_str(), "ab");
  if (fp == NULL) {
    LOGE("can not open file '%s'", file.c_str());
    return 1;
  }
  for (auto& e : list) {
    if (fputs((to_event_string(e) + "\n").c_str(), fp) < 0) {
      LOGE("can not write file '%s'", file.c_str());
      fclose(fp);
      return 2;
    }
  }
  fflush(fp);
  fclose(fp);
  return 0;
}

static int Record(const Args& args) {
  remove(args.record_file.c_str());
  Input input(args.device);
  std::list<struct input_event> event_list;
  struct input_event event;
  int ret = 0;
  while (!interrupt) {
    ret = input.Wait(&event, 200);
    if (ret > 0) {
      event_list.push_back(event);
      continue;
    }
    if (ret == 0) {
      if (event_list.empty()) {
        continue;
      }
      if (0 != AppendFile(event_list, args.record_file)) {
        return 1;
      }
      LOGD("已保存%u个事件", event_list.size());
      event_list.clear();
      continue;
    }
    if (ret < 0) {
      return 2;
    }
  }
  AppendFile(event_list, args.record_file);
  return 0;
}

struct timeval operator-(struct timeval t1, struct timeval t2) {
  struct timeval ret;
  /* Perform the carry for the later subtraction by updating y. */
  if (t1.tv_usec < t2.tv_usec) {
    int nsec = (t2.tv_usec - t1.tv_usec) / 1000000 + 1;
    t2.tv_usec -= 1000000 * nsec;
    t2.tv_sec += nsec;
  }
  if (t1.tv_usec - t2.tv_usec > 1000000) {
    int nsec = (t2.tv_usec - t1.tv_usec) / 1000000;
    t2.tv_usec += 1000000 * nsec;
    t2.tv_sec -= nsec;
  }

  /* Compute the time remaining to wait.
   tv_usec is certainly positive. */
  ret.tv_sec = t1.tv_sec - t2.tv_sec;
  ret.tv_usec = t1.tv_usec - t2.tv_usec;
  return ret;
}

static int Playback(const std::string& file, int speed) {
  FILE* fp = fopen(file.c_str(), "rb");
  if (fp == NULL) {
    LOGE("can not open file '%s'", file.c_str());
    return 1;
  }
  Input input(args.device);
  char buf[256];
  fgets(buf, sizeof(buf), fp);
  struct timeval last = from_event_string(buf).time;
  fseek(fp, 0, SEEK_SET);
  int ret = 0;
  while (!feof(fp)) {
    fgets(buf, sizeof(buf), fp);
    struct input_event event = from_event_string(buf);
    struct timeval diff = event.time - last;
    if (diff.tv_sec != 0 || diff.tv_usec != 0) {
      usleep((diff.tv_sec * 1000000 + diff.tv_usec) / speed);
    }
    if (0 != input.Inject(event)) {
      ret = 2;
      break;
    }
    last = event.time;
  }
  fclose(fp);
  return ret;
}

static int Playback(const Args& args) {
  for (int64_t i = 0; i < args.number_of_times || args.number_of_times == -1; ++i) {
    if (0 != Playback(args.record_file, args.speed)) {
      break;
    }
    LOGD("已回放%lld/%d遍", i + 1, args.number_of_times);
  }
  return 0;
}

int main(int argc, char** argv) {
  using namespace clipp;

  auto value_file = value("file", args.record_file).doc("记录文件名");

  auto command_tap = (
    command("tap").doc("点击").set(args.selected_cmd, Command::Tap),
    value("x", args.p1.x),
    value("y", args.p1.y),
    option("-r") & value("number of times", args.number_of_times)
    .doc("重复次数, 默认 1次, -1表示无限次")
  );
  auto command_swipe = (
    command("swipe").doc("滑动").set(args.selected_cmd, Command::Swipe),
    value("x1", args.p1.x),
    value("y1", args.p1.y),
    value("x2", args.p2.x),
    value("y2", args.p2.y)
  );
  auto command_playback = (
    command("play").doc("回放").set(args.selected_cmd, Command::Playback),
    value("file", args.record_file).doc("记录文件名"),
    option("-r") & value("number of times", args.number_of_times),
    option("-s") & value("speed", args.speed).doc("倍速 默认1")
  );
  auto command_record = (
    command("record").doc("录制").set(args.selected_cmd, Command::Record),
    value("file", args.record_file)
  );
  auto command_monkey = (
    command("monkey").doc("随机").set(args.selected_cmd, Command::Monkey),
    value("w", args.screen_width),
    value("h", args.screen_height),
    option("--no-swipe").set(args.no_swipe)
    .doc("没有滑动，只有点击")
  );

  auto cli = (
      value("device", args.device).doc("输入设备 例如 /dev/input/event0"),
      (command_tap | command_swipe | command_record | command_playback | command_monkey)
      );
  if (!parse(argc, argv, cli)) {
    std::cout << make_man_page(cli, "input");
    return 0;
  }

  signal(SIGINT, interrupt_handler);
  signal(SIGTERM, interrupt_handler);

  switch (args.selected_cmd) {
    case Command::Tap:
    {
      Input input(args.device);
      for (int i = 0; i < args.number_of_times; ++i) {
        input.Tap(args.p1.x, args.p1.y);
      }
    }
      break;
    case Command::Swipe:
    {
      Input input(args.device);
      for (int i = 0; i < args.number_of_times; ++i) {
        input.Swipe(args.p1.x, args.p1.y, args.p2.x, args.p2.y);
      }
    }
      break;
    case Command::Record:
      Record(args);
      break;
    case Command::Playback:
      if (args.speed <= 0) {
        LOGE("invalid speed");
        args.speed = 1;
      }
      Playback(args);
      break;
    case Command::Monkey:
      Monkey(args);
      break;
    default:
      std::cout << make_man_page(cli, "input");
      break;
  }
  return 0;
}
